# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .._common import _reproject_dispatcher
from ..utils import parse_input_data, parse_output_projection
from ._core import _reproject_adaptive_2d

__all__ = ["reproject_adaptive"]


def reproject_adaptive(
    input_data,
    output_projection,
    shape_out=None,
    hdu_in=0,
    center_jacobian=False,
    despike_jacobian=False,
    roundtrip_coords=True,
    conserve_flux=False,
    kernel="gaussian",
    kernel_width=1.3,
    sample_region_width=4,
    boundary_mode="strict",
    boundary_fill_value=0,
    boundary_ignore_threshold=0.5,
    x_cyclic=False,
    y_cyclic=False,
    bad_value_mode="strict",
    bad_fill_value=0,
    output_array=None,
    output_footprint=None,
    return_footprint=True,
    block_size=None,
    parallel=False,
    return_type=None,
    dask_method=None,
):
    """
    Reproject a 2D array from one WCS to another using the DeForest (2004)
    adaptive, anti-aliased resampling algorithm, with optional flux
    conservation. This algorithm smoothly transitions between filtered
    interpolation and spatial averaging, depending on the scaling applied by
    the transformation at each output location.

    Parameters
    ----------
    input_data : object
        The input data to reproject. This can be:

            * The name of a FITS file as a `str` or a `pathlib.Path` object
            * An `~astropy.io.fits.HDUList` object
            * An image HDU object such as a `~astropy.io.fits.PrimaryHDU`,
              `~astropy.io.fits.ImageHDU`, or `~astropy.io.fits.CompImageHDU`
              instance
            * A tuple where the first element is a `~numpy.ndarray` and the
              second element is either a
              `~astropy.wcs.wcsapi.BaseLowLevelWCS`,
              `~astropy.wcs.wcsapi.BaseHighLevelWCS`, or a
              `~astropy.io.fits.Header` object
            * An `~astropy.nddata.NDData` object from which the ``.data`` and
              ``.wcs`` attributes will be used as the input data.
            * The name of a PNG or JPEG file with AVM metadata

        If the data array contains more dimensions than are described by the
        input header or WCS, the extra dimensions (assumed to be the first
        dimensions) are taken to represent multiple images with the same
        coordinate information. The coordinate transformation will be computed
        once and then each image will be reprojected, offering a speedup over
        reprojecting each image individually.
    output_projection : `~astropy.wcs.wcsapi.BaseLowLevelWCS` or `~astropy.wcs.wcsapi.BaseHighLevelWCS` or `~astropy.io.fits.Header`
        The output projection, which can be either a
        `~astropy.wcs.wcsapi.BaseLowLevelWCS`,
        `~astropy.wcs.wcsapi.BaseHighLevelWCS`, or a `~astropy.io.fits.Header`
        instance.
    shape_out : tuple, optional
        If ``output_projection`` is a WCS instance, the shape of the output
        data should be specified separately.
    hdu_in : int or str, optional
        If ``input_data`` is a FITS file or an `~astropy.io.fits.HDUList`
        instance, specifies the HDU to use.
    center_jacobian : bool
        A Jacobian matrix is calculated, representing
        d(input image coordinate) / d(output image coordinate),
        a local linearization of the coordinate transformation. When this flag
        is ``True``, the Jacobian is calculated at pixel grid points by
        calculating the transformation at locations offset by half a pixel.
        This is more accurate but carries the cost of tripling the number of
        coordinate transforms done by this routine. This is recommended if your
        coordinate transform varies significantly and non-smoothly between
        output pixels. When ``False``, the Jacobian is calculated using
        pixel-grid-point transforms, which produces Jacobian values at
        locations between pixel grid points, and nearby Jacobian values are
        averaged to produce values at the pixel grid points. This is more
        efficient, and the loss of accuracy is extremely small for
        transformations that vary smoothly between pixels. Defaults to
        ``False``.
    despike_jacobian : bool
        Whether to despike the computed Jacobian values. In some situations
        (e.g. an all-sky map, with a wrap point in the longitude), extremely
        large Jacobian values may be computed which are artifacts of the
        coordinate system definition, rather than reflecting the actual nature
        of the coordinate transformation. This may result in a band of ``nan``
        pixels in the output image. In these situations, if the actual
        transformation is approximately constant in the region of these
        artifacts, this option should be enabled. If enabled, the typical
        magnitude (distance from the determinant) of the Jacobian matrix,
        ``Jmag2 = sum_j sum_i (J_ij**2)``, is computed for each pixel and
        compared to the 25th percentile of that value in the local 3x3
        neighborhood (i.e. the third-lowest value). If it exceeds that
        percentile value by more than 10 times, the Jacobian matrix is deemed
        to be "spiking" and it is replaced by the average of the non-spiking
        values in the 3x3 neighborhood.
    roundtrip_coords : bool
        Whether to verify that coordinate transformations are defined in both
        directions.
    conserve_flux : bool
        Whether to rescale output pixel values so flux is conserved.
    kernel : str
        The averaging kernel to use. Allowed values are 'Hann' and 'Gaussian'.
        Case-insensitive. The Gaussian kernel produces better photometric
        accuracy and stronger anti-aliasing at the cost of some blurring (on
        the scale of a few pixels). If not specified, the Gaussain kernel is
        used by default.
    kernel_width : double
        The width of the kernel in pixels, measuring to +/- 1 sigma for the
        Gaussian window. Does not apply to the Hann window. Reducing this width
        may introduce photometric errors or leave input pixels under-sampled,
        while increasing it may improve the degree of anti-aliasing but will
        increase blurring of the output image. If this width is changed from
        the default, a proportional change should be made to the value of
        sample_region_width to maintain an equivalent degree of photometric
        accuracy.
    sample_region_width : double
        The width in pixels of the output-image region which, when transformed
        to the input plane, defines the region to be sampled for each output
        pixel. Used only for the Gaussian kernel, which otherwise has infinite
        extent. This value sets a trade-off between accuracy and computation
        time, with better accuracy at higher values. The default value of 4,
        with the default kernel width, should limit the most extreme errors to
        less than one percent. Higher values will offer even more photometric
        accuracy.
    boundary_mode : str
        How to handle when the sampling region includes regions outside the
        bounds of the input image. The default is ``strict``. Allowed values
        are:

            * ``strict`` --- Output pixels will be NaN if any input sample
              falls outside the input image.
            * ``constant`` --- Samples outside the input image are replaced by
              a constant value, set with the ``boundary_fill_value`` argument.
              Output values become NaN if there are no valid input samples.
            * ``grid-constant`` --- Samples outside the input image are
              replaced by a constant value, set with the
              ``boundary_fill_value`` argument. Output values will be
              ``boundary_fill_value`` if there are no valid input samples.
            * ``ignore`` --- Samples outside the input image are simply
              ignored, contributing neither to the output value nor the
              sum-of-weights normalization.
            * ``ignore_threshold`` --- Acts as ``ignore``, unless the total
              weight of the ignored samples exceeds a set fraction of the total
              weight across the entire sampling region, set by the
              ``boundary_ignore_threshold`` argument. In that case, acts as
              ``strict``.
            * ``nearest`` --- Samples outside the input image are replaced by
              the nearest in-bounds input pixel.

    boundary_fill_value : double
        The constant value used by the ``constant`` boundary mode.
    boundary_ignore_threshold : double
        The threshold used by the ``ignore_threshold`` boundary mode. Should be
        a value between 0 and 1, representing a fraction of the total weight
        across the sampling region.
    x_cyclic, y_cyclic : bool
        Indicates that the x or y axis of the input image should be treated as
        cyclic or periodic. Overrides the boundary mode for that axis, so that
        out-of-bounds samples wrap to the other side of the image.
    bad_value_mode : str
        How to handle values of ``nan`` and ``inf`` in the input data. The
        default is ``strct``. Allowed values are:

            * ``strict`` --- Values of ``nan`` or ``inf`` in the input data are
              propagated to every output value which samples them.
            * ``ignore`` --- When a sampled input value is ``nan`` or ``inf``,
              that input pixel is ignored (affected neither the accumulated sum
              of weighted samples nor the accumulated sum of weights).
            * ``constant`` --- Input values of ``nan`` and ``inf`` are replaced
              with a constant value, set via the ``bad_fill_value`` argument.

    bad_fill_value : double
        The constant value used by the ``constant`` bad-value mode.
    output_array : None or `~numpy.ndarray`
        An array in which to store the reprojected data.  This can be any numpy
        array including a memory map, which may be helpful when dealing with
        extremely large files.
    output_footprint : `~numpy.ndarray`, optional
        An array in which to store the footprint of reprojected data.  This can be
        any numpy array including a memory map, which may be helpful when dealing with
        extremely large files.
    return_footprint : bool
        Whether to return the footprint in addition to the output array.
    block_size : tuple or 'auto', optional
        The size of blocks in terms of output array pixels that each block will handle
        reprojecting. Extending out from (0,0) coords positively, block sizes
        are clamped to output space edges when a block would extend past edge.
        Specifying ``'auto'`` means that reprojection will be done in blocks with
        the block size automatically determined. If ``block_size`` is not
        specified or set to `None`, the reprojection will not be carried out in
        blocks.
    parallel : bool or int or str, optional
        If `True`, the reprojection is carried out in parallel, and if a
        positive integer, this specifies the number of threads to use.
        The reprojection will be parallelized over output array blocks specified
        by ``block_size`` (if the block size is not set, it will be determined
        automatically). To use the currently active dask scheduler (e.g.
        dask.distributed), set this to ``'current-scheduler'``.
    return_type : {'numpy', 'dask'}, optional
        Whether to return numpy or dask arrays.
    dask_method : {'memmap', 'none'}, optional
        Method to use when input array is a dask array. The methods are:
            * ``'memmap'``: write out the entire input dask array to a temporary
              memory-mapped array. This requires enough disk space to store
              the entire input array, but should avoid accidentally loading
              the entire array into memory.
            * ``'none'``: load the dask array into memory as needed. This may
              result in the entire array being loaded into memory. However,
              this can be efficient under two conditions: if the array easily
              fits into memory (as this will then be faster than ``'memmap'``),
              and when the data contains more dimensions than the input WCS and
              the block_size is chosen to iterate over the extra dimensions.

    Returns
    -------
    array_new : `~numpy.ndarray`
        The reprojected array.
    footprint : `~numpy.ndarray`
        Footprint of the input array in the output array. Values of 0 indicate
        no coverage or valid values in the input image, while values of 1
        indicate valid values.
    """

    # TODO: add support for output_array

    array_in, wcs_in = parse_input_data(input_data, hdu_in=hdu_in)
    wcs_out, shape_out = parse_output_projection(
        output_projection, shape_in=array_in.shape, shape_out=shape_out
    )

    return _reproject_dispatcher(
        _reproject_adaptive_2d,
        array_in=array_in,
        wcs_in=wcs_in,
        wcs_out=wcs_out,
        shape_out=shape_out,
        array_out=output_array,
        parallel=parallel,
        block_size=block_size,
        return_footprint=return_footprint,
        output_footprint=output_footprint,
        reproject_func_kwargs=dict(
            center_jacobian=center_jacobian,
            despike_jacobian=despike_jacobian,
            roundtrip_coords=roundtrip_coords,
            conserve_flux=conserve_flux,
            kernel=kernel,
            kernel_width=kernel_width,
            sample_region_width=sample_region_width,
            boundary_mode=boundary_mode,
            boundary_fill_value=boundary_fill_value,
            boundary_ignore_threshold=boundary_ignore_threshold,
            x_cyclic=x_cyclic,
            y_cyclic=y_cyclic,
            bad_value_mode=bad_value_mode,
            bad_fill_value=bad_fill_value,
        ),
        return_type=return_type,
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import itertools
import os
from copy import deepcopy

import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS

from .. import reproject_adaptive, reproject_exact, reproject_interp

# TODO: add reference comparisons

DATA = os.path.join(os.path.dirname(__file__), "data")

ALL_MODES = (
    "nearest-neighbor",
    "bilinear",
    "biquadratic",
    "bicubic",
    "flux-conserving",
    "adaptive-hann",
    "adaptive-gaussian",
)

ALL_DTYPES = []
for endian in ("<", ">"):
    for kind in ("u", "i", "f"):
        for size in ("1", "2", "4", "8"):
            if kind == "f" and size == "1":
                continue
            ALL_DTYPES.append(np.dtype(endian + kind + size))


@pytest.fixture(
    params=[reproject_interp, reproject_adaptive, reproject_exact],
    ids=["interp", "adaptive", "exact"],
)
def reproject_function(request):
    return request.param


class TestReproject:
    def setup_method(self, method):
        self.header_in = fits.Header.fromtextfile(get_pkg_data_filename("data/gc_ga.hdr"))
        self.header_out = fits.Header.fromtextfile(get_pkg_data_filename("data/gc_eq.hdr"))

        self.header_out_size = self.header_out.copy()
        self.header_out_size["NAXIS"] = 2
        self.header_out_size["NAXIS1"] = 600
        self.header_out_size["NAXIS2"] = 550

        self.array_in = np.ones((700, 690))

        self.hdu_in = fits.ImageHDU(self.array_in, self.header_in)

        self.wcs_in = WCS(self.header_in)
        self.wcs_out = WCS(self.header_out)
        self.shape_out = (600, 550)

    @pytest.mark.filterwarnings("ignore::FutureWarning")
    def test_hdu_header(self, reproject_function):
        with pytest.raises(ValueError) as exc:
            reproject_function(self.hdu_in, self.header_out)
        assert exc.value.args[0] == (
            "Need to specify shape since output header "
            "does not contain complete shape information"
        )

        reproject_interp(self.hdu_in, self.header_out_size)

    @pytest.mark.filterwarnings("ignore::FutureWarning")
    def test_hdu_wcs(self, reproject_function):
        reproject_function(self.hdu_in, self.wcs_out, shape_out=self.shape_out)

    @pytest.mark.filterwarnings("ignore::FutureWarning")
    def test_array_wcs_header(self, reproject_function):
        with pytest.raises(ValueError) as exc:
            reproject_function((self.array_in, self.wcs_in), self.header_out)
        assert exc.value.args[0] == (
            "Need to specify shape since output header "
            "does not contain complete shape information"
        )

        reproject_function((self.array_in, self.wcs_in), self.header_out_size)

    @pytest.mark.filterwarnings("ignore::FutureWarning")
    def test_array_wcs_wcs(self, reproject_function):
        reproject_function((self.array_in, self.wcs_in), self.wcs_out, shape_out=self.shape_out)

    @pytest.mark.filterwarnings("ignore::FutureWarning")
    def test_array_header_header(self, reproject_function):
        reproject_function((self.array_in, self.header_in), self.header_out_size)

    @pytest.mark.filterwarnings("ignore::FutureWarning")
    def test_return_footprint(self, reproject_function):
        array = reproject_function(
            self.hdu_in, self.wcs_out, shape_out=self.shape_out, return_footprint=False
        )
        assert isinstance(array, np.ndarray)


INPUT_HDR = """
WCSAXES =                    2 / Number of coordinate axes
CRPIX1  =                  0.5 / Pixel coordinate of reference point
CRPIX2  =                  0.5 / Pixel coordinate of reference point
CDELT1  =         -0.001666666 / [deg] Coordinate increment at reference point
CDELT2  =          0.001666666 / [deg] Coordinate increment at reference point
CUNIT1  = 'deg'                / Units of coordinate increment and value
CUNIT2  = 'deg'                / Units of coordinate increment and value
CTYPE1  = 'GLON-CAR'           / galactic longitude, plate caree projection
CTYPE2  = 'GLAT-CAR'           / galactic latitude, plate caree projection
CRVAL1  =                  0.0 / [deg] Coordinate value at reference point
CRVAL2  =                  0.0 / [deg] Coordinate value at reference point
LONPOLE =                  0.0 / [deg] Native longitude of celestial pole
LATPOLE =                 90.0 / [deg] Native latitude of celestial pole
"""


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.parametrize("projection_type, dtype", list(itertools.product(ALL_MODES, ALL_DTYPES)))
def test_surface_brightness(projection_type, dtype):
    header_in = fits.Header.fromstring(INPUT_HDR, sep="\n")
    header_in["NAXIS"] = 2
    header_in["NAXIS1"] = 10
    header_in["NAXIS2"] = 10

    header_out = header_in.copy()

    header_out["CDELT1"] /= 2
    header_out["CDELT2"] /= 2
    header_out["NAXIS1"] *= 2
    header_out["NAXIS2"] *= 2

    data_in = np.ones((10, 10), dtype=dtype)

    if projection_type == "flux-conserving":
        data_out, footprint = reproject_exact((data_in, header_in), header_out)
    elif projection_type.startswith("adaptive"):
        data_out, footprint = reproject_adaptive(
            (data_in, header_in),
            header_out,
            kernel=projection_type.split("-", 1)[1],
            boundary_mode="ignore",
        )
    else:
        data_out, footprint = reproject_interp(
            (data_in, header_in), header_out, order=projection_type
        )

    assert data_out.shape == (20, 20)

    # Here we check that the values are still 1 despite the change in
    # resolution, which demonstrates that we are preserving surface
    # brightness.
    np.testing.assert_allclose(data_out, 1)


IDENTITY_TEST_HDR = """
NAXIS   =                    2 / Number of coordinate axes
NAXIS1  =                    5
NAXIS2  =                   10
CRPIX1  =  8540.80750619681    / Pixel coordinate of reference point
CRPIX2  =   4108.61481031444   / Pixel coordinate of reference point
CUNIT1  = 'deg'                / Units of coordinate increment and value
CUNIT2  = 'deg'                / Units of coordinate increment and value
CTYPE1  = 'RA---TAN'           / galactic longitude, plate caree projection
CTYPE2  = 'DEC--TAN'           / galactic latitude, plate caree projection
CRVAL1  =  282.582253365684    / [deg] Coordinate value at reference point
CRVAL2  =   -5.80644283270032  / [deg] Coordinate value at reference point
CD1_1   = 5.92448829959494E-05 / WCS transform matrix element
CD2_1   = -2.5640855053008E-08 / WCS transform matrix element
CD1_2   = 1.30443859769625E-08 / WCS transform matrix element
CD2_2   = 5.92479929826863E-05 / WCS transform matrix element
"""


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.parametrize("projection_type", ALL_MODES)
def test_identity_projection(projection_type):
    """Sanity check: identical input & output headers should preserve image."""
    header_in = fits.Header.fromstring(IDENTITY_TEST_HDR, sep="\n")
    data_in = np.random.rand(header_in["NAXIS2"], header_in["NAXIS1"])
    if projection_type == "flux-conserving":
        data_out, footprint = reproject_exact((data_in, header_in), header_in)
    elif projection_type.startswith("adaptive"):
        data_out, footprint = reproject_adaptive(
            (data_in, header_in),
            header_in,
            kernel=projection_type.split("-", 1)[1],
            boundary_mode="ignore",
        )
    else:
        data_out, footprint = reproject_interp(
            (data_in, header_in), header_in, order=projection_type
        )
    # When reprojecting with an identical input and output header,
    # we may expect the input and output data to be similar,
    # and the footprint values to be ~ones.
    expected_footprint = np.ones((header_in["NAXIS2"], header_in["NAXIS1"]))
    np.testing.assert_allclose(footprint, expected_footprint)
    if projection_type != "adaptive-gaussian":
        np.testing.assert_allclose(data_in, data_out, rtol=1e-6)
    else:
        # The Gaussian kernel in the adaptive algorithm blurs the image
        assert np.all(data_in != data_out)


@pytest.mark.parametrize(
    "reproject_function", [reproject_interp, reproject_adaptive, reproject_exact]
)
def test_dimensions_checks(reproject_function):
    header_in = fits.Header.fromtextfile(
        get_pkg_data_filename("data/gc_eq.hdr", package="reproject.tests")
    )

    header_out = header_in.copy()
    w_in = WCS(header_in)
    w_out = WCS(header_out)

    array_in = np.zeros((2, 3, 4, 5))

    # Create mis-matches between the input and output shapes

    with pytest.raises(
        ValueError, match="Number of dimensions in input and output data should match"
    ):
        array_out, footprint = reproject_function(
            (array_in, w_in), w_out, shape_out=[2, 3, 4, 5, 6]
        )

    with pytest.raises(
        ValueError, match="Number of dimensions in input and output data should match"
    ):
        array_out, footprint = reproject_function(
            (array_in[0], w_in), w_out, shape_out=[3, 4, 5, 6]
        )

    # Give fewer array dimensions than WCS dimensions

    with pytest.raises(ValueError, match="Too few dimensions in input data"):
        array_out, footprint = reproject_function(
            (array_in[0, 0, 0], w_in), w_out, shape_out=[4, 5, 6]
        )

    with pytest.raises(ValueError, match="Too few dimensions in shape_out"):
        array_out, footprint = reproject_function(
            (array_in[0], w_in),
            w_out,
            shape_out=[
                5,
            ],
        )

    with pytest.raises(ValueError, match="Dimensions to be looped over must match exactly"):
        array_out, footprint = reproject_function((array_in, w_in), w_out, shape_out=[3, 3, 4, 5])


@pytest.mark.remote_data
@pytest.mark.filterwarnings("ignore::astropy.wcs.wcs.FITSFixedWarning")
@pytest.mark.parametrize(
    "reproject_function", [reproject_interp, reproject_adaptive, reproject_exact]
)
@pytest.mark.parametrize("scheduler", ("synchronous", "processes", "threads"))
# @pytest.mark.parametrize("wcs_type", ("astropy.wcs", "gwcs"))
@pytest.mark.parametrize("wcs_type", ("gwcs",))
def test_dask_schedulers(reproject_function, scheduler, wcs_type):
    # Regression test for issues with the multi-threaded scheduler

    kwargs = {}

    if wcs_type == "astropy.wcs":
        input_data = fits.open(get_pkg_data_filename("galactic_center/gc_2mass_k.fits"))[0]
        hdu_out = fits.open(get_pkg_data_filename("galactic_center/gc_msx_e.fits"))[0]
        wcs_out = WCS(hdu_out.header)
        shape_out = hdu_out.data.shape
    elif wcs_type == "gwcs":
        pytest.importorskip("sunpy", minversion="6.0.1")
        asdf = pytest.importorskip("asdf")
        if reproject_function == reproject_exact:
            pytest.skip()
        aia = asdf.open(os.path.join(DATA, "aia_171_level1.asdf"))
        input_data = (aia["data"][...], aia["wcs"])
        wcs_out = deepcopy(aia["wcs"])
        wcs_out.forward_transform.offset_0 = -60.3123 * u.pix
        wcs_out.forward_transform.offset_1 = -61.9422 * u.pix
        shape_out = aia["data"].shape
        kwargs["roundtrip_coords"] = False

    array1 = reproject_function(
        input_data,
        wcs_out,
        shape_out=shape_out,
        return_footprint=False,
        **kwargs,
    )

    array2 = reproject_function(
        input_data,
        wcs_out,
        shape_out=shape_out,
        return_footprint=False,
        return_type="dask",
        block_size=(100, 100),
        **kwargs,
    )
    array2 = array2.compute(scheduler=scheduler)

    np.testing.assert_allclose(array1, array2, equal_nan=True)



import logging
import os
import tempfile
import uuid

import dask
import dask.array as da
import numpy as np
from astropy.wcs import WCS
from astropy.wcs.wcsapi import BaseHighLevelWCS, SlicedLowLevelWCS
from astropy.wcs.wcsapi.high_level_wcs_wrapper import HighLevelWCSWrapper
from dask import delayed

from ._array_utils import ArrayWrapper
from .utils import _dask_to_numpy_memmap

__all__ = ["_reproject_dispatcher"]


class _ArrayContainer:
    # When we set up as_delayed_memmap_path, if we pass a dask array to it,
    # dask will actually compute the array before we get to the code inside
    # as_delayed_memmap_path, so as a workaround we wrap any array we
    # pass in using _ArrayContainer to make sure dask doesn't try and be smart.
    def __init__(self, array):
        self._array = array


@delayed(pure=True)
def as_delayed_memmap_path(array, tmp_dir):

    # Extract array from _ArrayContainer
    if isinstance(array, _ArrayContainer):
        array = array._array
    else:
        raise TypeError("Expected _ArrayContainer in as_delayed_memmap_path")

    logger = logging.getLogger(__name__)
    if isinstance(array, da.core.Array):
        logger.info("Computing input dask array to Numpy memory-mapped array")
        array_path, _ = _dask_to_numpy_memmap(array, tmp_dir)
        logger.info(f"Numpy memory-mapped array is now at {array_path}")
    else:
        array_path = os.path.join(tmp_dir, f"{uuid.uuid4()}.npy")
        array_memmapped = np.memmap(
            array_path,
            dtype=float,
            shape=array.shape,
            mode="w+",
        )
        array_memmapped[:] = array[:]

    return array_path


def _reproject_dispatcher(
    reproject_func,
    *,
    array_in,
    wcs_in,
    shape_out,
    wcs_out,
    block_size=None,
    array_out=None,
    return_footprint=True,
    output_footprint=None,
    parallel=True,
    reproject_func_kwargs=None,
    return_type=None,
    dask_method=None,
):
    """
    Main function that handles either calling the core algorithms directly or
    parallelizing or operating in chunks, using dask.

    Parameters
    ----------
    reproject_func
        One the existing reproject functions implementing a reprojection algorithm
        that that will be used be used to perform reprojection
    array_in : `numpy.ndarray` or `dask.array.Array`
        Numpy or dask input array
    wcs_in: `~astropy.wcs.wcsapi.BaseHighLevelWCS`
        Input data WCS
    shape_out: tuple
        Target shape
    wcs_out: `~astropy.wcs.WCS`
        Target WCS
    block_size: tuple or 'auto', optional
        The size of blocks in terms of output array pixels that each block will handle
        reprojecting. Extending out from (0,0) coords positively, block sizes
        are clamped to output space edges when a block would extend past edge.
        Specifying ``'auto'`` means that reprojection will be done in blocks with
        the block size automatically determined. If ``block_size`` is not
        specified or set to `None`, the reprojection will not be carried out in
        blocks.
    array_out : `~numpy.ndarray`, optional
        An array in which to store the reprojected data.  This can be any numpy
        array including a memory map, which may be helpful when dealing with
        extremely large files.
    return_footprint : bool, optional
        Whether to return the footprint in addition to the output array.
    output_footprint : `~numpy.ndarray`, optional
        An array in which to store the footprint of reprojected data.  This can be
        any numpy array including a memory map, which may be helpful when dealing with
        extremely large files.
    parallel : bool or int or str, optional
        If `True`, the reprojection is carried out in parallel, and if a
        positive integer, this specifies the number of threads to use.
        The reprojection will be parallelized over output array blocks specified
        by ``block_size`` (if the block size is not set, it will be determined
        automatically). To use the currently active dask scheduler (e.g.
        dask.distributed), set this to ``'current-scheduler'``.
    reproject_func_kwargs : dict, optional
        Keyword arguments to pass through to ``reproject_func``
    return_type : {'numpy', 'dask' }, optional
        Whether to return numpy or dask arrays.
    dask_method : {'memmap', 'none'}, optional
        Method to use when input array is a dask array. The methods are:
            * ``'memmap'``: write out the entire input dask array to a temporary
              memory-mapped array. This requires enough disk space to store
              the entire input array, but should avoid accidentally loading
              the entire array into memory.
            * ``'none'``: load the dask array into memory as needed. This may
              result in the entire array being loaded into memory. However,
              this can be efficient under two conditions: if the array easily
              fits into memory (as this will then be faster than ``'memmap'``),
              and when the data contains more dimensions than the input WCS and
              the block_size is chosen to iterate over the extra dimensions.
    """

    logger = logging.getLogger(__name__)

    if return_type is None:
        return_type = "numpy"
    elif return_type not in ("numpy", "dask"):
        raise ValueError("return_type should be set to 'numpy' or 'dask'")

    if dask_method is None:
        dask_method = "memmap"
    elif dask_method not in ("memmap", "none"):
        raise ValueError("dask_method should be set to 'memmap' or 'none'")

    if reproject_func_kwargs is None:
        reproject_func_kwargs = {}

    # We set up a global temporary directory since this will be used e.g. to
    # store memory mapped Numpy arrays and zarr arrays.

    with tempfile.TemporaryDirectory() as local_tmp_dir:
        if array_out is None:
            if return_type != "dask":
                array_out = np.zeros(shape_out, dtype=float)
        elif array_out.shape != tuple(shape_out):
            raise ValueError(
                f"Output array shape {array_out.shape} should match " f"shape_out={shape_out}"
            )
        elif (array_out.dtype.kind, array_out.dtype.itemsize) != (
            array_in.dtype.kind,
            array_in.dtype.itemsize,
        ):
            # Note that here we don't care if the endians don't match
            raise ValueError(
                f"Output array dtype {array_out.dtype} should match "
                f"input array dtype ({array_in.dtype})"
            )

        # If neither parallel nor blocked reprojection are requested, we simply
        # call the underlying core reproject function with the full arrays.

        if block_size is None and parallel is False:
            # If a dask array was passed as input, we first convert this to a
            # Numpy memory mapped array

            if return_type == "dask":
                raise ValueError(
                    "Output cannot be returned as dask arrays "
                    "when parallel=False and no block size has "
                    "been specified"
                )

            if isinstance(array_in, da.core.Array) and dask_method == "memmap":
                logger.info("Computing input dask array to Numpy memory-mapped array")
                array_path, array_in = _dask_to_numpy_memmap(array_in, local_tmp_dir)
                logger.info(f"Numpy memory-mapped array is now at {array_path}")

            logger.debug(f"Calling {reproject_func.__name__} in non-dask mode")

            try:
                return reproject_func(
                    array_in,
                    wcs_in,
                    wcs_out,
                    shape_out=shape_out,
                    array_out=array_out,
                    return_footprint=return_footprint,
                    output_footprint=output_footprint,
                    **reproject_func_kwargs,
                )
            finally:
                # Clean up reference to numpy memmap
                array_in = None

        # Determine whether any broadcasting is taking place. This is when the
        # input/output WCS have fewer dimensions that the data, and any preceding
        # dimensions can be assumed to be independent of the WCS. At this point
        # shape_out will be the full size of the output array as this is updated
        # in parse_output_projection, even if shape_out was originally passed in as
        # the shape of a single image.
        broadcasting = wcs_in.low_level_wcs.pixel_n_dim < len(shape_out)

        logger.info(f"Broadcasting is {'' if broadcasting else 'not '}being used")

        # Check block size and determine whether block size indicates we should
        # parallelize over broadcasted dimension. The logic is as follows: if
        # the block size and output shape are the same size, then either the
        # block size should match the output shape along the broadcasted
        # dimensions or along the non-broadcasted dimensions. If it matches the
        # non-broadcasted dimensions we can parallelize over the broadcasted
        # dimensions. If the block size does not match the output shape, we
        # don't make any assumptions for now and assume a single chunk in the
        # missing dimensions.
        broadcasted_parallelization = False
        if broadcasting and block_size is not None and block_size != "auto":
            if len(block_size) == len(shape_out):
                if (
                    block_size[-wcs_in.low_level_wcs.pixel_n_dim :]
                    == shape_out[-wcs_in.low_level_wcs.pixel_n_dim :]
                ):
                    broadcasted_parallelization = True
                    block_size = (
                        block_size[: -wcs_in.low_level_wcs.pixel_n_dim]
                        + (-1,) * wcs_in.low_level_wcs.pixel_n_dim
                    )
                else:
                    for i in range(len(shape_out) - wcs_in.low_level_wcs.pixel_n_dim):
                        if block_size[i] != -1 and block_size[i] != shape_out[i]:
                            raise ValueError(
                                "block shape should either match output data shape along broadcasted dimension or non-broadcasted dimensions"
                            )
            elif len(block_size) < len(shape_out):
                block_size = [-1] * (len(shape_out) - len(block_size)) + list(block_size)
            else:
                raise ValueError(
                    f"block_size {len(block_size)} cannot have more elements "
                    f"than the dimensionality of the output ({len(shape_out)})"
                )

            # TODO: check for shape_out not matching shape_in along broadcasted dimensions

        logger.info(
            f"{'P' if broadcasted_parallelization else 'Not p'}arallelizing along "
            f"broadcasted dimension ({block_size=}, {shape_out=})"
        )

        if output_footprint is None and return_footprint:
            output_footprint = np.zeros(shape_out, dtype=float)

        shape_in = array_in.shape

        def reproject_single_block(a, array_or_path, block_info=None):

            if (
                a.ndim == 0
                or block_info is None
                or block_info == []
                or (isinstance(block_info, np.ndarray) and block_info.tolist() == [])
            ):
                return np.array([a, a])

            if isinstance(array_or_path, str) and array_or_path == "from-dict":
                array_or_path = dask_arrays["array"]

            # The WCS class from astropy is not thread-safe, see e.g.
            # https://github.com/astropy/astropy/issues/16244
            # https://github.com/astropy/astropy/issues/16245
            # To work around these issues, we make sure we do a deep copy of
            # the WCS object in here when using FITS WCS. This is a very fast
            # operation (<0.1ms) so should not be a concern in terms of
            # performance. We only need to do this for FITS WCS.

            wcs_in_cp = wcs_in.deepcopy() if isinstance(wcs_in, WCS) else wcs_in
            wcs_out_cp = wcs_out.deepcopy() if isinstance(wcs_out, WCS) else wcs_out

            slices = [
                slice(*x) for x in block_info[None]["array-location"][-wcs_out_cp.pixel_n_dim :]
            ]

            if isinstance(wcs_out, BaseHighLevelWCS):
                low_level_wcs = SlicedLowLevelWCS(wcs_out_cp.low_level_wcs, slices=slices)
            else:
                low_level_wcs = SlicedLowLevelWCS(wcs_out_cp, slices=slices)

            wcs_out_sub = HighLevelWCSWrapper(low_level_wcs)

            if isinstance(array_or_path, tuple):
                array_in = np.memmap(array_or_path[0], **array_or_path[1], mode="r")
            elif isinstance(array_or_path, str):
                array_in = np.memmap(array_or_path, dtype=float, shape=shape_in, mode="r")
            else:
                array_in = array_or_path

            if array_or_path is None:
                raise RuntimeError("array_or_path is not set")

            shape_out = block_info[None]["chunk-shape"][1:]

            array, footprint = reproject_func(
                array_in,
                wcs_in_cp,
                wcs_out_sub,
                shape_out=shape_out,
                array_out=np.zeros(shape_out),
                **reproject_func_kwargs,
            )

            return np.array([array, footprint])

        if broadcasted_parallelization:

            array_out_dask = da.empty(shape_out, chunks=block_size)
            if isinstance(array_in, da.core.Array):
                if array_in.chunksize != block_size:
                    logger.info(
                        f"Rechunking input dask array as chunks ({array_in.chunksize}) "
                        "do not match block size ({block_size})"
                    )
                    array_in = array_in.rechunk(block_size)
            else:

                array_in = da.asarray(
                    ArrayWrapper(array_in), name=str(uuid.uuid4()), chunks=block_size
                )

            result = da.map_blocks(
                reproject_single_block,
                array_out_dask,
                array_in,
                dtype="<f8",
                new_axis=0,
                chunks=((2,),) + array_out_dask.chunks,
            )

        else:

            # As we use the synchronous or threads scheduler, we don't need to worry about
            # the data getting copied, so if the data is already a Numpy array (including
            # a memory-mapped array) then we don't need to do anything special. However,
            # if the input array is a dask array, we should convert it to a Numpy
            # memory-mapped array so that it can be used by the various reprojection
            # functions (which don't internally work with dask arrays).

            if isinstance(array_in, np.memmap) and array_in.flags.c_contiguous:
                array_in_or_path = array_in.filename, {
                    "dtype": array_in.dtype,
                    "shape": array_in.shape,
                    "offset": array_in.offset,
                }
            elif isinstance(array_in, da.core.Array) or return_type == "dask":
                if dask_method == "memmap":
                    if return_type == "dask":
                        # We should use a temporary directory that will persist beyond
                        # the call to the reproject function.
                        tmp_dir = tempfile.mkdtemp()
                    else:
                        tmp_dir = local_tmp_dir
                    array_in_or_path = as_delayed_memmap_path(_ArrayContainer(array_in), tmp_dir)
                else:
                    dask_arrays = {"array": array_in}
                    array_in_or_path = "from-dict"
            else:
                # Here we could set array_in_or_path to array_in_path if it has
                # been set previously, but in synchronous and threaded mode it is
                # better to simply pass a reference to the memmap array itself to
                # avoid having to load the memmap inside each
                # reproject_single_block call.
                array_in_or_path = array_in

            if block_size is not None and block_size != "auto":
                array_out_dask = da.empty(shape_out, chunks=block_size)
            else:
                if broadcasting:
                    chunks = (-1,) * (len(shape_out) - wcs_in.low_level_wcs.pixel_n_dim)
                    chunks += ("auto",) * wcs_in.low_level_wcs.pixel_n_dim
                    rechunk_kwargs = {"chunks": chunks}
                else:
                    rechunk_kwargs = {}
                array_out_dask = da.empty(shape_out)
                array_out_dask = array_out_dask.rechunk(
                    block_size_limit=64 * 1024**2, **rechunk_kwargs
                )

            logger.info("Setting up output dask array with map_blocks")

            result = da.map_blocks(
                reproject_single_block,
                array_out_dask,
                array_in_or_path,
                dtype="<f8",
                new_axis=0,
                chunks=(2,) + array_out_dask.chunksize,
            )

        # Ensure that there are no more references to Numpy memmaps
        array_in = None
        array_in_or_path = None

        # Truncate extra elements
        result = result[tuple([slice(None)] + [slice(s) for s in shape_out])]

        if return_type == "dask":
            if return_footprint:
                return result[0], result[1]
            else:
                return result[0]

        # We now convert the dask arrays back to Numpy arrays

        if parallel:
            # As discussed in https://github.com/dask/dask/issues/9556, da.store
            # will not work well in parallel mode when the destination is a
            # Numpy array. Instead, in this case we save the dask array to a zarr
            # array on disk which can be done in parallel, and re-load it as a dask
            # array. We can then use da.store in the next step using the
            # 'synchronous' scheduler since that is I/O limited so does not need
            # to be done in parallel.

            zarr_path = os.path.join(local_tmp_dir, f"{uuid.uuid4()}.zarr")

            logger.info(f"Computing output array directly to zarr array at {zarr_path}")

            if parallel == "current-scheduler":
                # Just use whatever is the current active scheduler, which can
                # be used for e.g. dask.distributed
                result.to_zarr(zarr_path)
            else:
                if isinstance(parallel, bool):
                    workers = {}
                else:
                    if parallel > 0:
                        workers = {"num_workers": parallel}
                    else:
                        raise ValueError(
                            "The number of processors to use must be strictly positive"
                        )

                with dask.config.set(scheduler="threads", **workers):
                    result.to_zarr(zarr_path)

            result = da.from_zarr(zarr_path)

        logger.info("Copying output zarr array into output Numpy arrays")

        if return_footprint:
            da.store(
                [result[0], result[1]],
                [array_out, output_footprint],
                compute=True,
                scheduler="synchronous",
            )
            output = array_out, output_footprint
        else:
            da.store(
                result[0],
                array_out,
                compute=True,
                scheduler="synchronous",
            )
            output = array_out

    if return_footprint:
        return output[0], output[1]
    else:
        return output



import os
import tempfile
import uuid
from pathlib import Path

import astropy.nddata
import dask.array as da
import numpy as np
from astropy.io import fits
from astropy.io.fits import CompImageHDU, HDUList, Header, ImageHDU, PrimaryHDU
from astropy.wcs import WCS
from astropy.wcs.wcsapi import BaseHighLevelWCS, BaseLowLevelWCS
from astropy.wcs.wcsapi.high_level_wcs_wrapper import HighLevelWCSWrapper
from PIL import Image
from pyavm import AVM

__all__ = [
    "parse_input_data",
    "parse_input_shape",
    "parse_input_weights",
    "parse_output_projection",
    "as_transparent_rgb",
]


def _dask_to_numpy_memmap(dask_array, tmp_dir):
    """
    Given a dask array, write it out to disk and load it again as a Numpy
    memmap array.
    """

    # Sometimes compute() has to be called twice to return a Numpy array,
    # so we need to check here if this is the case and call the first compute()
    if isinstance(dask_array[(slice(0, 0),) * dask_array.ndim].compute(), da.Array):
        dask_array = dask_array.compute()

    # Cast the dask array to regular float for two reasons - first, zarr 3.0.0
    # and later doesn't support big-endian arrays, and also we need to anyway
    # make a native float memory mapped array below.
    dask_array = dask_array.astype("<f8", copy=False)

    # NOTE: here we use a new TemporaryDirectory context manager for the zarr
    # array because we can remove the temporary directory straight away after
    # converting the input to a Numpy memory mapped array.

    with tempfile.TemporaryDirectory() as zarr_tmp:
        # First compute and store the dask array to zarr using whatever
        # the default scheduler is at this point
        try:
            dask_array.to_zarr(zarr_tmp)
        except ValueError:
            dask_array.rechunk().to_zarr(zarr_tmp)

        # Load the array back to dask
        zarr_array = da.from_zarr(zarr_tmp)

        # Then store this in the Numpy memmap array (schedulers other than
        # synchronous don't work well when writing to a Numpy array)

        memmap_path = os.path.join(tmp_dir, f"{uuid.uuid4()}.npy")

        memmapped_array = np.memmap(
            memmap_path,
            dtype=float,
            shape=zarr_array.shape,
            mode="w+",
        )

        da.store(
            zarr_array,
            memmapped_array,
            compute=True,
            scheduler="synchronous",
        )

    return memmap_path, memmapped_array


def hdu_to_numpy_memmap(hdu):
    """
    Given an HDU object, return a regular Numpy memmap rather than a Numpy
    array backed by a memmapped buffer as returned by astropy.
    """

    if (
        getattr(hdu, "_orig_bscale", 1) != 1
        or getattr(hdu, "_orig_bzero", 0) != 0
        or hdu.header.get("BSCALE", 1) != 1
        or hdu.header.get("BZERO", 0) != 0
        or hdu.fileinfo() is None
        or hdu._data_replaced
        or hdu.fileinfo()["file"].compression is not None
    ):
        return hdu.data

    return np.memmap(
        hdu.fileinfo()["file"].name,
        mode="r",
        dtype=hdu.data.dtype.newbyteorder(">"),
        shape=hdu.data.shape,
        offset=hdu.fileinfo()["datLoc"],
    )


def parse_input_data(input_data, hdu_in=None, source_hdul=None):
    """
    Parse input data to return a Numpy array and WCS object.
    """

    if isinstance(input_data, str | Path):
        if is_png(input_data) or is_jpeg(input_data):
            data = np.array(Image.open(input_data)).transpose(2, 0, 1)[:, ::-1]
            wcs = AVM.from_image(input_data).to_wcs()
            return data, wcs
        else:
            with fits.open(input_data) as hdul:
                return parse_input_data(hdul, hdu_in=hdu_in)
    elif isinstance(input_data, HDUList):
        if hdu_in is None:
            if len(input_data) > 1:
                raise ValueError(
                    "More than one HDU is present, please specify "
                    "HDU to use with ``hdu_in=`` option"
                )
            else:
                hdu_in = 0
        return parse_input_data(input_data[hdu_in], source_hdul=input_data)
    elif isinstance(input_data, PrimaryHDU | ImageHDU) and not isinstance(input_data, CompImageHDU):
        return (hdu_to_numpy_memmap(input_data), WCS(input_data.header, fobj=source_hdul))
    elif isinstance(input_data, CompImageHDU):
        return (input_data.data, WCS(input_data.header, fobj=source_hdul))
    elif isinstance(input_data, tuple) and isinstance(input_data[0], np.ndarray | da.core.Array):
        if isinstance(input_data[1], Header):
            return input_data[0], WCS(input_data[1])
        else:
            if isinstance(input_data[1], BaseHighLevelWCS):
                return input_data[0], input_data[1]
            else:
                return input_data[0], HighLevelWCSWrapper(input_data[1])
    elif (
        isinstance(input_data, BaseHighLevelWCS)
        and input_data.low_level_wcs.array_shape is not None
    ):
        return input_data.array_shape, input_data
    elif isinstance(input_data, BaseLowLevelWCS) and input_data.array_shape is not None:
        return input_data.array_shape, HighLevelWCSWrapper(input_data)
    elif isinstance(input_data, astropy.nddata.NDDataBase):
        return input_data.data, input_data.wcs
    else:
        raise TypeError(
            "input_data should either be an HDU object or a tuple "
            "of (array, WCS) or (array, Header)"
        )


def parse_input_shape(input_shape, hdu_in=None):
    """
    Parse input shape information to return an array shape tuple and WCS object.
    """

    if isinstance(input_shape, str | Path):
        if is_png(input_shape) or is_jpeg(input_shape):
            shape = np.array(Image.open(input_shape)).transpose(2, 0, 1).shape
            wcs = AVM.from_image(input_shape).to_wcs()
            return shape, wcs
        else:
            with fits.open(input_shape) as hdulist:
                return parse_input_shape(hdulist, hdu_in=hdu_in)
    elif isinstance(input_shape, HDUList):
        if hdu_in is None:
            if len(input_shape) > 1:
                raise ValueError(
                    "More than one HDU is present, please specify "
                    "HDU to use with ``hdu_in=`` option"
                )
            else:
                hdu_in = 0
        return parse_input_shape(input_shape[hdu_in])
    elif isinstance(input_shape, PrimaryHDU | ImageHDU | CompImageHDU):
        return input_shape.shape, WCS(input_shape.header)
    elif isinstance(input_shape, tuple) and isinstance(input_shape[0], np.ndarray | da.core.Array):
        if isinstance(input_shape[1], Header):
            return input_shape[0].shape, WCS(input_shape[1])
        else:
            if isinstance(input_shape[1], BaseHighLevelWCS):
                return input_shape[0].shape, input_shape[1]
            else:
                return input_shape[0].shape, HighLevelWCSWrapper(input_shape[1])
    elif isinstance(input_shape, tuple) and isinstance(input_shape[0], tuple):
        if isinstance(input_shape[1], Header):
            return input_shape[0], WCS(input_shape[1])
        else:
            if isinstance(input_shape[1], BaseHighLevelWCS):
                return input_shape
            else:
                return input_shape[0], HighLevelWCSWrapper(input_shape[1])
    elif (
        isinstance(input_shape, BaseHighLevelWCS)
        and input_shape.low_level_wcs.array_shape is not None
    ):
        return input_shape.low_level_wcs.array_shape, input_shape
    elif isinstance(input_shape, BaseLowLevelWCS) and input_shape.array_shape is not None:
        return input_shape.array_shape, HighLevelWCSWrapper(input_shape)
    elif isinstance(input_shape, astropy.nddata.NDDataBase):
        return input_shape.data.shape, input_shape.wcs
    else:
        raise TypeError(
            "input_shape should either be an HDU object or a tuple "
            "of (array-or-shape, WCS) or (array-or-shape, Header)"
        )


def parse_input_weights(input_weights, hdu_weights=None):
    """
    Parse input weights to return a Numpy array.
    """

    if isinstance(input_weights, str):
        return parse_input_data(fits.open(input_weights), hdu_in=hdu_weights)
    elif isinstance(input_weights, HDUList):
        if hdu_weights is None:
            if len(input_weights) > 1:
                raise ValueError(
                    "More than one HDU is present, please specify "
                    "HDU to use with ``hdu_weights=`` option"
                )
            else:
                hdu_weights = 0
        return parse_input_data(input_weights[hdu_weights])
    elif isinstance(input_weights, PrimaryHDU | ImageHDU | CompImageHDU):
        if "CTYPE1" in input_weights.header:
            # all valid WCSes have CTYPE1 specified, at least
            ww = WCS(input_weights.header)
        else:
            ww = None
        return input_weights.data, ww
    elif isinstance(input_weights, np.ndarray):
        return input_weights, None
    else:
        raise TypeError("input_weights should either be an HDU object or a Numpy array")


def parse_output_projection(output_projection, shape_in=None, shape_out=None, output_array=None):
    if shape_out is None:
        if output_array is not None:
            shape_out = output_array.shape
    elif shape_out is not None and output_array is not None:
        if shape_out != output_array.shape:
            raise ValueError("shape_out does not match shape of output_array")

    if isinstance(output_projection, Header):
        wcs_out = WCS(output_projection)
        try:
            shape_out = [
                output_projection[f"NAXIS{i + 1}"] for i in range(output_projection["NAXIS"])
            ][::-1]
        except KeyError:
            if shape_out is None:
                raise ValueError(
                    "Need to specify shape since output header "
                    "does not contain complete shape information"
                ) from None
    elif isinstance(output_projection, BaseLowLevelWCS | BaseHighLevelWCS):
        if isinstance(output_projection, BaseLowLevelWCS) and not isinstance(
            output_projection, BaseHighLevelWCS
        ):
            wcs_out = HighLevelWCSWrapper(output_projection)
        else:
            wcs_out = output_projection
        if shape_out is None:
            if wcs_out.low_level_wcs.array_shape is not None:
                shape_out = wcs_out.low_level_wcs.array_shape
            else:
                raise ValueError(
                    "Need to specify shape_out when specifying output_projection as WCS object"
                )
    elif isinstance(output_projection, str):
        hdu_list = fits.open(output_projection)
        shape_out = hdu_list[0].data.shape
        header = hdu_list[0].header
        wcs_out = WCS(header)
        hdu_list.close()
    else:
        raise TypeError("output_projection should either be a Header, a WCS object, or a filename")

    if len(shape_out) == 0:
        raise ValueError("The shape of the output image should not be an empty tuple")

    if (
        shape_in is not None
        and len(shape_out) < len(shape_in)
        and len(shape_out) == wcs_out.low_level_wcs.pixel_n_dim
    ):
        # Add the broadcast dimensions to the output shape, which does not
        # currently have any broadcast dims
        shape_out = (*shape_in[: -len(shape_out)], *shape_out)
    return wcs_out, tuple(shape_out)


def is_png(filename):
    with open(filename, "rb") as f:
        return f.read(8) == b"\x89\x50\x4e\x47\x0d\x0a\x1a\x0a"


def is_jpeg(filename):
    with open(filename, "rb") as f:
        return f.read(3) == b"\xff\xd8\xff"


def as_transparent_rgb(data, alpha=None):
    """
    Convert a 3D Numpy array to a PIL object.

    This takes care of swapping the order of the axes, and can apply an additional
    transparency layer. This also converts any NaNs to transparency.

    Parameters
    ----------
    data : `numpy.ndarray`
        Input array with shape ``(3, ny, nx)`` or ``(4, ny, nx)``
    alpha : `numpy.ndarray`
        Alpha layer to apply to the image in addition to any pre-existing alpha
        layer and transparency originating from NaN values. This should have
        a shape ``(ny, nx)``
    """

    if data.ndim != 3:
        raise ValueError("Data needs to be three-dimensional to return RGB image")

    if data.shape[0] not in (3, 4):
        raise ValueError("Data should have shape (3, ny, nx) or (4, ny, nx)")

    array = np.zeros((4,) + data.shape[1:], dtype=np.uint8)

    if alpha is None:

        alpha = np.ones(data.shape[1:])

    else:

        if alpha.ndim != 2:
            raise ValueError("alpha needs to be two-dimensional")

        if alpha.shape != data.shape[1:]:
            raise ValueError(
                "alpha layer shape {alpha.shape} does not match data spatial shape {data.shape[1:]}"
            )

        alpha = alpha.copy()

    alpha[np.any(np.isnan(data), axis=0)] = 0

    data = np.nan_to_num(data)

    if data.shape[0] == 3:
        array[:3] = data
        array[3] = 255
    else:
        array[...] = data

    array[3] = array[3] * alpha

    return Image.fromarray(array.transpose(1, 2, 0)[::-1])



from collections.abc import Iterable

import astropy.modeling.models as m
import astropy.units as u
import gwcs.coordinate_frames as cf
from astropy.modeling import CompoundModel, Model
from gwcs import WCS
from numpy.typing import ArrayLike


def generate_celestial_transform(
    crpix: Iterable[float] | u.Quantity,
    cdelt: Iterable[float] | u.Quantity,
    pc: ArrayLike | u.Quantity,
    crval: Iterable[float] | u.Quantity,
    lon_pole: float | u.Quantity = None,
    projection: Model = None,
) -> CompoundModel:
    """
    Create a simple celestial transform from FITS like parameters.

    Supports unitful or unitless parameters, but if any parameters have units
    all must have units, if parameters are unitless they are assumed to be in
    degrees.

    Parameters
    ----------
    crpix
        The reference pixel (a length two array).
    crval
        The world coordinate at the reference pixel (a length two array).
    cdelt
        The sample interval along the pixel axis
    pc
        The rotation matrix for the affine transform. If specifying parameters
        with units this should have celestial (``u.deg``) units.
    lon_pole
        The longitude of the celestial pole, defaults to 180 degrees.
    projection
        The map projection to use, defaults to ``TAN``.

    Notes
    -----

    This function has not been tested with more complex projections. Ensure
    that your lon_pole is correct for your projection.
    """
    projection = m.Pix2Sky_TAN() if projection is None else None
    spatial_unit = None
    if hasattr(crval[0], "unit"):
        spatial_unit = crval[0].unit
    # TODO: Note this assumption is only valid for certain projections.
    if lon_pole is None:
        lon_pole = 180
    if spatial_unit is not None:
        # Lon pole should always have the units of degrees
        lon_pole = u.Quantity(lon_pole, unit=u.deg)

    # Make translation unitful if all parameters have units
    translation = (0, 0)
    if spatial_unit is not None:
        translation *= pc.unit
        # If we have units then we need to convert all things to Quantity
        # as they might be Parameter classes
        crpix = u.Quantity(crpix)
        cdelt = u.Quantity(cdelt)
        crval = u.Quantity(crval)
        lon_pole = u.Quantity(lon_pole)
        pc = u.Quantity(pc)

    shift = m.Shift(-crpix[0]) & m.Shift(-crpix[1])
    scale = m.Multiply(cdelt[0]) & m.Multiply(cdelt[1])
    rot = m.AffineTransformation2D(pc, translation=translation)
    skyrot = m.RotateNative2Celestial(crval[0], crval[1], lon_pole)
    return shift | rot | scale | projection | skyrot


def generate_asdf(input_file="aia_171_level1.fits", output_file="aia_171_level1.asdf"):
    # Put imports for optional or not dependencies here
    import asdf
    import sunpy.map

    aia_map = sunpy.map.Map(input_file)

    transform = generate_celestial_transform(
        crpix=aia_map.reference_pixel,
        cdelt=aia_map.scale,
        pc=aia_map.rotation_matrix * u.pix,
        crval=aia_map.wcs.wcs.crval * u.deg,
    )

    input_frame = cf.Frame2D()
    output_frame = cf.CelestialFrame(
        reference_frame=aia_map.coordinate_frame,
        unit=(u.arcsec, u.arcsec),
        axes_names=("Helioprojective Longitude", "Helioprojective Latitude"),
    )

    aia_gwcs = WCS(
        forward_transform=transform,
        input_frame=input_frame,
        output_frame=output_frame,
    )

    tree = {
        "data": aia_map.data,
        "meta": dict(aia_map.meta),
        "wcs": aia_gwcs,
    }

    af = asdf.AsdfFile(tree)
    af.write_to(output_file)


if __name__ == "__main__":
    generate_asdf()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .._common import _reproject_dispatcher
from .._wcs_utils import has_celestial
from ..utils import parse_input_data, parse_output_projection
from ._core import _reproject_celestial

__all__ = ["reproject_exact"]


def reproject_exact(
    input_data,
    output_projection,
    shape_out=None,
    hdu_in=0,
    output_array=None,
    output_footprint=None,
    return_footprint=True,
    block_size=None,
    parallel=False,
    return_type=None,
    dask_method=None,
):
    """
    Reproject data to a new projection using flux-conserving spherical
    polygon intersection (this is the slowest algorithm).

    Parameters
    ----------
    input_data : object
        The input data to reproject. This can be:

            * The name of a FITS file as a `str` or a `pathlib.Path` object
            * An `~astropy.io.fits.HDUList` object
            * An image HDU object such as a `~astropy.io.fits.PrimaryHDU`,
              `~astropy.io.fits.ImageHDU`, or `~astropy.io.fits.CompImageHDU`
              instance
            * A tuple where the first element is a `~numpy.ndarray` and the
              second element is either a
              `~astropy.wcs.wcsapi.BaseLowLevelWCS`,
              `~astropy.wcs.wcsapi.BaseHighLevelWCS`, or a
              `~astropy.io.fits.Header` object
            * An `~astropy.nddata.NDData` object from which the ``.data`` and
              ``.wcs`` attributes will be used as the input data.
            * The name of a PNG or JPEG file with AVM metadata

        If the data array contains more dimensions than are described by the
        input header or WCS, the extra dimensions (assumed to be the first
        dimensions) are taken to represent multiple images with the same
        coordinate information. The coordinate transformation will be computed
        once and then each image will be reprojected, offering a speedup over
        reprojecting each image individually.
    output_projection : `~astropy.wcs.wcsapi.BaseLowLevelWCS` or `~astropy.wcs.wcsapi.BaseHighLevelWCS` or `~astropy.io.fits.Header`
        The output projection, which can be either a
        `~astropy.wcs.wcsapi.BaseLowLevelWCS`,
        `~astropy.wcs.wcsapi.BaseHighLevelWCS`, or a `~astropy.io.fits.Header`
        instance.
    shape_out : tuple, optional
        If ``output_projection`` is a WCS instance, the shape of the output
        data should be specified separately.
    hdu_in : int or str, optional
        If ``input_data`` is a FITS file or an `~astropy.io.fits.HDUList`
        instance, specifies the HDU to use.
    output_array : None or `~numpy.ndarray`
        An array in which to store the reprojected data.  This can be any numpy
        array including a memory map, which may be helpful when dealing with
        extremely large files.
    output_footprint : `~numpy.ndarray`, optional
        An array in which to store the footprint of reprojected data.  This can be
        any numpy array including a memory map, which may be helpful when dealing with
        extremely large files.
    return_footprint : bool
        Whether to return the footprint in addition to the output array.
    block_size : tuple or 'auto', optional
        The size of blocks in terms of output array pixels that each block will handle
        reprojecting. Extending out from (0,0) coords positively, block sizes
        are clamped to output space edges when a block would extend past edge.
        Specifying ``'auto'`` means that reprojection will be done in blocks with
        the block size automatically determined. If ``block_size`` is not
        specified or set to `None`, the reprojection will not be carried out in
        blocks.
    parallel : bool or int or str, optional
        If `True`, the reprojection is carried out in parallel, and if a
        positive integer, this specifies the number of threads to use.
        The reprojection will be parallelized over output array blocks specified
        by ``block_size`` (if the block size is not set, it will be determined
        automatically). To use the currently active dask scheduler (e.g.
        dask.distributed), set this to ``'current-scheduler'``.
    return_type : {'numpy', 'dask'}, optional
        Whether to return numpy or dask arrays
    dask_method : {'memmap', 'none'}, optional
        Method to use when input array is a dask array. The methods are:
            * ``'memmap'``: write out the entire input dask array to a temporary
              memory-mapped array. This requires enough disk space to store
              the entire input array, but should avoid accidentally loading
              the entire array into memory.
            * ``'none'``: load the dask array into memory as needed. This may
              result in the entire array being loaded into memory. However,
              this can be efficient under two conditions: if the array easily
              fits into memory (as this will then be faster than ``'memmap'``),
              and when the data contains more dimensions than the input WCS and
              the block_size is chosen to iterate over the extra dimensions.

    Returns
    -------
    array_new : `~numpy.ndarray`
        The reprojected array.
    footprint : `~numpy.ndarray`
        Footprint of the input array in the output array. Values of 0 indicate
        no coverage or valid values in the input image, while values of 1
        indicate valid values. Intermediate values indicate partial coverage.
    """

    array_in, wcs_in = parse_input_data(input_data, hdu_in=hdu_in)
    wcs_out, shape_out = parse_output_projection(
        output_projection, shape_in=array_in.shape, shape_out=shape_out
    )

    if has_celestial(wcs_in) and wcs_in.pixel_n_dim == 2 and wcs_in.world_n_dim == 2:
        return _reproject_dispatcher(
            _reproject_celestial,
            array_in=array_in,
            wcs_in=wcs_in,
            wcs_out=wcs_out,
            shape_out=shape_out,
            array_out=output_array,
            parallel=parallel,
            block_size=block_size,
            return_footprint=return_footprint,
            output_footprint=output_footprint,
            return_type=return_type,
        )
    else:
        raise NotImplementedError(
            "Currently only data with a 2-d celestial "
            "WCS can be reprojected using flux-conserving algorithm"
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import operator

__all__ = ["ReprojectedArraySubset"]


class ReprojectedArraySubset:
    # The aim of this class is to represent a subset of an array and
    # footprint extracted (or meant to represent extracted) versions
    # from larger arrays and footprints.

    # NOTE: we can't use Cutout2D here because it's much more convenient
    # to work with position being the lower left corner of the cutout
    # rather than the center, which is not well defined for even-sized
    # cutouts.

    def __init__(self, array, footprint, bounds):
        self.array = array
        self.footprint = footprint
        self.bounds = bounds

    def __repr__(self):
        bounds_str = "[" + ",".join(f"{imin}:{imax}" for (imin, imax) in self.bounds) + "]"
        return f"<ReprojectedArraySubset at {bounds_str}>"

    @property
    def view_in_original_array(self):
        return tuple([slice(imin, imax) for (imin, imax) in self.bounds])

    @property
    def shape(self):
        return tuple((imax - imin) for (imin, imax) in self.bounds)

    def overlaps(self, other):
        # Note that the use of <= or >= instead of < and > is due to
        # the fact that the max values are exclusive (so +1 above the
        # last value).
        if len(self.bounds) != len(other.bounds):
            raise ValueError(
                f"Mismatch in number of dimensions, expected "
                f"{len(self.bounds)} dimensions and got {len(other.bounds)}"
            )
        for (imin, imax), (imin_other, imax_other) in zip(self.bounds, other.bounds, strict=False):
            if imax <= imin_other or imax_other <= imin:
                return False
        return True

    def __add__(self, other):
        return self._operation(other, operator.add)

    def __sub__(self, other):
        return self._operation(other, operator.sub)

    def __mul__(self, other):
        return self._operation(other, operator.mul)

    def __truediv__(self, other):
        return self._operation(other, operator.truediv)

    def _operation(self, other, op):
        if len(self.bounds) != len(other.bounds):
            raise ValueError(
                f"Mismatch in number of dimensions, expected "
                f"{len(self.bounds)} dimensions and got {len(other.bounds)}"
            )

        # Determine cutout parameters for overlap region

        overlap_bounds = []
        self_slices = []
        other_slices = []
        for (imin, imax), (imin_other, imax_other) in zip(self.bounds, other.bounds, strict=False):
            imin_overlap = max(imin, imin_other)
            imax_overlap = min(imax, imax_other)
            if imax_overlap < imin_overlap:
                imax_overlap = imin_overlap
            overlap_bounds.append((imin_overlap, imax_overlap))
            self_slices.append(slice(imin_overlap - imin, imax_overlap - imin))
            other_slices.append(slice(imin_overlap - imin_other, imax_overlap - imin_other))

        self_slices = tuple(self_slices)

        self_array = self.array[self_slices]
        self_footprint = self.footprint[self_slices]

        other_slices = tuple(other_slices)

        other_array = other.array[other_slices]
        other_footprint = other.footprint[other_slices]

        # Carry out operator and store result in ReprojectedArraySubset

        array = op(self_array, other_array)
        footprint = (self_footprint > 0) & (other_footprint > 0)

        return ReprojectedArraySubset(array, footprint, overlap_bounds)



import numpy as np

from ._overlap import _compute_overlap

__all__ = ["compute_overlap"]


def compute_overlap(ilon, ilat, olon, olat):
    """Compute the overlap between two 'pixels' in spherical coordinates.

    Parameters
    ----------
    ilon : np.ndarray with shape (N, 4)
        The longitudes (in radians) defining the four corners of the input pixel
    ilat : np.ndarray with shape (N, 4)
        The latitudes (in radians) defining the four corners of the input pixel
    olon : np.ndarray with shape (N, 4)
        The longitudes (in radians) defining the four corners of the output pixel
    olat : np.ndarray with shape (N, 4)
        The latitudes (in radians) defining the four corners of the output pixel

    Returns
    -------
    overlap : np.ndarray of length N
        Pixel overlap solid angle in steradians
    area_ratio : np.ndarray of length N
        TODO
    """
    ilon = np.ascontiguousarray(ilon, dtype=np.float64)
    ilat = np.ascontiguousarray(ilat, dtype=np.float64)
    olon = np.ascontiguousarray(olon, dtype=np.float64)
    olat = np.ascontiguousarray(olat, dtype=np.float64)

    return _compute_overlap(ilon, ilat, olon, olat)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os
import sys
import tempfile
import uuid
from logging import getLogger

import numpy as np
from astropy.wcs import WCS
from astropy.wcs.utils import pixel_to_pixel
from astropy.wcs.wcsapi import SlicedLowLevelWCS

from .._array_utils import iterate_chunks, sample_array_edges
from ..interpolation._core import _validate_wcs
from ..utils import parse_input_data, parse_input_weights, parse_output_projection
from ._background import determine_offset_matrix, solve_corrections_sgd
from ._subset_array import ReprojectedArraySubset

__all__ = ["reproject_and_coadd"]


IS_WIN = sys.platform == "win32"


def _noop(iterable):
    return iterable


def reproject_and_coadd(
    input_data,
    output_projection,
    shape_out=None,
    input_weights=None,
    hdu_in=None,
    hdu_weights=None,
    reproject_function=None,
    combine_function="mean",
    match_background=False,
    background_reference=None,
    output_array=None,
    output_footprint=None,
    block_sizes=None,
    progress_bar=None,
    blank_pixel_value=0,
    intermediate_memmap=False,
    **kwargs,
):
    """
    Given a set of input data, reproject and co-add these to a single
    final image.

    Parameters
    ----------
    input_data : iterable
        One or more input datasets to reproject and co-add. This should be an
        iterable containing one entry for each dataset, where a single dataset
        is one of:

            * The name of a FITS file
            * An `~astropy.io.fits.HDUList` object
            * An image HDU object such as a `~astropy.io.fits.PrimaryHDU`,
              `~astropy.io.fits.ImageHDU`, or `~astropy.io.fits.CompImageHDU`
              instance
            * A tuple where the first element is an `~numpy.ndarray` and the
              second element is either a `~astropy.wcs.WCS` or a
              `~astropy.io.fits.Header` object
            * An `~astropy.nddata.NDData` object from which the ``.data`` and
              ``.wcs`` attributes will be used as the input data.

    output_projection : `~astropy.wcs.wcsapi.BaseLowLevelWCS` or `~astropy.wcs.wcsapi.BaseHighLevelWCS` or `~astropy.io.fits.Header`
        The output projection, which can be either a
        `~astropy.wcs.wcsapi.BaseLowLevelWCS`,
        `~astropy.wcs.wcsapi.BaseHighLevelWCS`, or a `~astropy.io.fits.Header`
        instance.
    shape_out : tuple, optional
        If ``output_projection`` is a WCS instance, the shape of the output
        data should be specified separately.
    input_weights : iterable
        If specified, this should be an iterable with the same length as
        ``input_data``, where each item is one of:

            * The name of a FITS file
            * An `~astropy.io.fits.HDUList` object
            * An image HDU object such as a `~astropy.io.fits.PrimaryHDU`,
              `~astropy.io.fits.ImageHDU`, or `~astropy.io.fits.CompImageHDU`
              instance
            * An `~numpy.ndarray` array

    hdu_in : int or str, optional
        If one or more items in ``input_data`` is a FITS file or an
        `~astropy.io.fits.HDUList` instance, specifies the HDU to use.
    hdu_weights : int or str, optional
        If one or more items in ``input_weights`` is a FITS file or an
        `~astropy.io.fits.HDUList` instance, specifies the HDU to use.
    reproject_function : callable
        The function to use for the reprojection.
    combine_function : { 'mean', 'sum', 'first', 'last', 'min', 'max' }
        The type of function to use for combining the values into the final
        image. For 'first' and 'last', respectively, the reprojected images are
        simply overlaid on top of each other. With respect to the order of the
        input images in ``input_data``, either the first or the last image to
        cover a region of overlap determines the output data for that region.
    match_background : bool
        Whether to match the backgrounds of the images.
    background_reference : `None` or `int`
        If `None`, the background matching will make it so that the average of
        the corrections for all images is zero. If an integer, this specifies
        the index of the image to use as a reference.
    output_array : array or None
        The final output array.  Specify this if you already have an
        appropriately-shaped array to store the data in.  Must match shape
        specified with ``shape_out`` or derived from the output
        projection.
    output_footprint : array or None
        The final output footprint array.  Specify this if you already have an
        appropriately-shaped array to store the data in.  Must match shape
        specified with ``shape_out`` or derived from the output projection.
    block_sizes : list of tuples or None
        The block size to use for each dataset.  Could also be a single tuple
        if you want the same block size for all data sets.
    progress_bar : callable, optional
        If specified, use this as a progress_bar to track loop iterations over
        data sets.
    blank_pixel_value : float, optional
        Value to use for areas of the resulting mosaic that do not have input
        data.
    intermediate_memmap : bool, optional
        If `True`, use `numpy.memmap` to store intermediate output arrays for
        reprojected data.

    **kwargs
        Keyword arguments to be passed to the reprojection function.

    Returns
    -------
    array : `~numpy.ndarray`
        The co-added array.
    footprint : `~numpy.ndarray`
        Footprint of the co-added array. Values of 0 indicate no coverage or
        valid values in the input image, while values of 1 indicate valid
        values.
    """

    # TODO: add support for saving intermediate files to disk to avoid blowing
    # up memory usage. We could probably still have references to array
    # objects, but we'd just make sure these were memory mapped

    logger = getLogger(__name__)

    # Validate inputs

    if combine_function not in ("mean", "sum", "first", "last", "min", "max"):
        raise ValueError("combine_function should be one of mean/sum/first/last/min/max")

    if reproject_function is None:
        raise ValueError(
            "reprojection function should be specified with the reproject_function argument"
        )

    if "block_size" in kwargs and kwargs["block_size"] is not None:
        if block_sizes is not None:
            raise ValueError("Cannot specify block_sizes= and block_size= at the same time")
        block_sizes = kwargs.pop("block_size")

    if progress_bar is None:
        progress_bar = _noop

    # Parse the output projection to avoid having to do it for each

    wcs_out, shape_out = parse_output_projection(output_projection, shape_out=shape_out)

    if output_array is None:
        output_array = np.zeros(shape_out)
    elif output_array.shape != shape_out:
        raise ValueError(
            "If you specify an output array, it must have a shape matching "
            f"the output shape {shape_out}"
        )

    if output_footprint is None:
        output_footprint = np.zeros(shape_out)
    elif output_footprint.shape != shape_out:
        raise ValueError(
            "If you specify an output footprint array, it must have a shape matching "
            f"the output shape {shape_out}"
        )

    logger.info(f"Output mosaic will have shape {shape_out}")

    # Define 'on-the-fly' mode: in the case where we don't need to match
    # the backgrounds and we are combining with 'mean' or 'sum', we don't
    # have to keep track of the intermediate arrays and can just modify
    # the output array on-the-fly
    on_the_fly = not match_background and combine_function in ("mean", "sum")

    on_the_fly_prefix = "Using" if on_the_fly else "Not using"
    logger.info(
        f"{on_the_fly_prefix} on-the-fly mode for adding individual reprojected images to output array"
    )

    # Start off by reprojecting individual images to the final projection

    if not on_the_fly:
        arrays = []

    with tempfile.TemporaryDirectory(ignore_cleanup_errors=IS_WIN) as local_tmp_dir:

        for idata in progress_bar(range(len(input_data))):

            logger.info(f"Processing input data {idata + 1} of {len(input_data)}")

            # We need to pre-parse the data here since we need to figure out how to
            # optimize/minimize the size of each output tile (see below).
            array_in, wcs_in = parse_input_data(input_data[idata], hdu_in=hdu_in)

            # We also get the weights map, if specified
            if input_weights is None:
                weights_in = None
            else:
                weights_in, weights_wcs = parse_input_weights(
                    input_weights[idata], hdu_weights=hdu_weights
                )
                if weights_wcs is None:
                    # if weights are passed as an array
                    weights_wcs = wcs_in
                else:
                    try:
                        _validate_wcs(weights_wcs, wcs_in, weights_in.shape, shape_out)
                    except ValueError:
                        # WCS is not valid (most likely, it is blank?)
                        weights_wcs = wcs_in
                if np.any(np.isnan(weights_in)):
                    weights_in = np.nan_to_num(weights_in)

            # Since we might be reprojecting small images into a large mosaic we
            # want to make sure that for each image we reproject to an array with
            # minimal footprint. We therefore find the pixel coordinates of the
            # edges of the initial image and transform this to pixel coordinates in
            # the final image to figure out the final WCS and shape to reproject to
            # for each tile. We strike a balance between transforming only the
            # input-image corners, which is fast but can cause clipping in cases of
            # significant distortion (when the edges of the input image become
            # convex in the output projection), and transforming every edge pixel,
            # which provides a lot of redundant information.

            edges = sample_array_edges(
                array_in.shape[-wcs_in.low_level_wcs.pixel_n_dim :], n_samples=11
            )[::-1]
            edges_out = pixel_to_pixel(wcs_in, wcs_out, *edges)[::-1]

            # Determine the cutout parameters

            # In some cases, images might not have valid coordinates in the corners,
            # such as all-sky images or full solar disk views. In this case we skip
            # this step and just use the full output WCS for reprojection.

            ndim_out = len(shape_out)

            # Determine how many extra broadcasted dimensions are present
            n_broadcasted = len(shape_out) - wcs_in.low_level_wcs.pixel_n_dim

            skip_data = False
            if np.any(np.isnan(edges_out)):
                bounds = list(zip([0] * ndim_out, shape_out, strict=False))
            else:
                bounds = []
                if n_broadcasted > 0:
                    for idim in range(n_broadcasted):
                        bounds.append((0, shape_out[idim]))
                for idim in range(len(edges_out)):
                    imin = max(0, int(np.floor(edges_out[idim].min() + 0.5)))
                    imax = min(
                        shape_out[n_broadcasted + idim], int(np.ceil(edges_out[idim].max() + 0.5))
                    )
                    bounds.append((imin, imax))
                    if imax <= imin:
                        skip_data = True
                        break

            if skip_data:
                logger.info(
                    "Skipping reprojection as no predicted overlap with final mosaic header"
                )
                continue

            slice_out = tuple([slice(imin, imax) for (imin, imax) in bounds[n_broadcasted:]])

            if isinstance(wcs_out, WCS):
                wcs_out_indiv = wcs_out[slice_out]
            else:
                wcs_out_indiv = SlicedLowLevelWCS(wcs_out.low_level_wcs, slice_out)

            shape_out_indiv = tuple([imax - imin for (imin, imax) in bounds])

            if block_sizes is not None:
                if len(block_sizes) == len(input_data) and len(block_sizes[idata]) == len(
                    shape_out
                ):
                    global_block_size = block_sizes[idata]
                else:
                    global_block_size = block_sizes
            else:
                global_block_size = None

            # If the block size matches the non-broadcasted shape of the final
            # cube, we need to update the non-broadcasted shape to then match
            # the subset size.
            if (
                global_block_size
                and global_block_size[-wcs_in.low_level_wcs.pixel_n_dim]
                == shape_out[-wcs_in.low_level_wcs.pixel_n_dim]
            ):
                block_size = global_block_size[:n_broadcasted] + shape_out_indiv[n_broadcasted:]
            else:
                block_size = global_block_size

            # TODO: optimize handling of weights by making reprojection functions
            # able to handle weights, and make the footprint become the combined
            # footprint + weight map

            if intermediate_memmap:

                array_path = os.path.join(local_tmp_dir, f"array_{uuid.uuid4()}.np")

                logger.info(
                    f"Creating memory-mapped array with shape {shape_out_indiv} at {array_path}"
                )

                array = np.memmap(
                    array_path,
                    shape=shape_out_indiv,
                    mode="w+",
                    dtype=array_in.dtype,
                )

                footprint_path = os.path.join(local_tmp_dir, f"footprint_{uuid.uuid4()}.np")

                logger.info(
                    f"Creating memory-mapped footprint with shape {shape_out_indiv} at {footprint_path}"
                )

                footprint = np.memmap(
                    footprint_path,
                    shape=shape_out_indiv,
                    mode="w+",
                    dtype=float,
                )

            else:

                array = footprint = None

            logger.info(f"Calling {reproject_function.__name__} with shape_out={shape_out_indiv}")

            array, footprint = reproject_function(
                (array_in, wcs_in),
                output_projection=wcs_out_indiv,
                shape_out=shape_out_indiv,
                hdu_in=hdu_in,
                output_array=array,
                output_footprint=footprint,
                block_size=block_size,
                **kwargs,
            )

            if weights_in is not None:

                if intermediate_memmap:

                    weights_path = os.path.join(local_tmp_dir, f"weights_{uuid.uuid4()}.np")

                    logger.info(
                        f"Creating memory-mapped weights with shape {shape_out_indiv} at {weights_path}"
                    )

                    weights = np.memmap(
                        weights_path,
                        shape=shape_out_indiv,
                        mode="w+",
                        dtype=float,
                    )

                else:

                    weights = None

                logger.info(
                    f"Calling {reproject_function.__name__} with shape_out={shape_out_indiv} for weights"
                )

                weights = reproject_function(
                    (weights_in, weights_wcs),
                    output_projection=wcs_out_indiv,
                    shape_out=shape_out_indiv,
                    hdu_in=hdu_in,
                    output_array=weights,
                    return_footprint=False,
                    **kwargs,
                )
                reset = np.isnan(array) | np.isnan(weights)
            else:
                reset = np.isnan(array)

            # For the purposes of mosaicking, we mask out NaN values from the array
            # and set the footprint to 0 at these locations.
            array[reset] = 0.0
            footprint[reset] = 0.0

            # Combine weights and footprint
            if weights_in is not None:
                weights[reset] = 0.0
                footprint *= weights

                if intermediate_memmap:
                    # Remove the reference to the memmap before trying to remove the file itself
                    logger.info("Removing memory-mapped weight array")
                    weights = None
                    try:
                        os.remove(weights_path)
                    except PermissionError:
                        pass

            array = ReprojectedArraySubset(array, footprint, bounds)

            # TODO: make sure we gracefully handle the case where the
            # output image is empty (due e.g. to no overlap).

            if on_the_fly:
                logger.info("Adding reprojected array to final array")
                # By default, values outside of the footprint are set to NaN
                # but we set these to 0 here to avoid getting NaNs in the
                # means/sums.
                array.array[array.footprint == 0] = 0
                output_footprint[array.view_in_original_array] += array.footprint
                # We now need to do output[view] += array * footprint but to avoid
                # the temporary array allocation from array * footprint we modify
                # array inplace, which we can do as the array will be discarded at
                # the end of the loop.
                array.array *= array.footprint
                output_array[array.view_in_original_array] += array.array

                if intermediate_memmap:
                    # Remove the references to the memmaps themesleves before
                    # trying to remove the files thermselves.
                    logger.info("Removing memory-mapped array and footprint arrays")
                    array = None
                    footprint = None
                    try:
                        os.remove(array_path)
                        os.remove(footprint_path)
                    except PermissionError:
                        pass

            else:
                logger.info("Adding reprojected array to list to combine later")
                arrays.append(array)

        # If requested, try and match the backgrounds.
        if match_background and len(arrays) > 1:
            logger.info("Match backgrounds")
            offset_matrix = determine_offset_matrix(arrays)
            corrections = solve_corrections_sgd(offset_matrix)
            if background_reference:
                corrections -= corrections[background_reference]
            for array, correction in zip(arrays, corrections, strict=True):
                array.array -= correction

        if combine_function in ("mean", "sum"):
            if match_background:
                logger.info(f"Combining reprojected arrays with function {combine_function}")
                # if we're not matching the background, this part has already been done
                for array in arrays:
                    # By default, values outside of the footprint are set to NaN
                    # but we set these to 0 here to avoid getting NaNs in the
                    # means/sums.
                    array.array[array.footprint == 0] = 0
                    output_array[array.view_in_original_array] += array.array * array.footprint
                    output_footprint[array.view_in_original_array] += array.footprint

            if combine_function == "mean":
                logger.info("Handle normalization of output array")
                with np.errstate(invalid="ignore"):
                    output_array /= output_footprint

        elif combine_function in ("first", "last", "min", "max"):
            logger.info(f"Combining reprojected arrays with function {combine_function}")
            if combine_function == "min":
                output_array[...] = np.inf
            elif combine_function == "max":
                output_array[...] = -np.inf

            for array in arrays:
                if combine_function == "first":
                    mask = output_footprint[array.view_in_original_array] == 0
                elif combine_function == "last":
                    mask = array.footprint > 0
                elif combine_function == "min":
                    mask = (array.footprint > 0) & (
                        array.array < output_array[array.view_in_original_array]
                    )
                elif combine_function == "max":
                    mask = (array.footprint > 0) & (
                        array.array > output_array[array.view_in_original_array]
                    )

                output_footprint[array.view_in_original_array] = np.where(
                    mask, array.footprint, output_footprint[array.view_in_original_array]
                )
                output_array[array.view_in_original_array] = np.where(
                    mask, array.array, output_array[array.view_in_original_array]
                )

        # Avoid keeping any references to the memory-mapped arrays so that the
        # files get cleaned up once we exit the context manager.
        if intermediate_memmap:
            array = None
            footprint = None
            arrays = []

    # We need to avoid potentially large memory allocation from output == 0 so
    # we operate in chunks.
    logger.info(f"Resetting invalid pixels to {blank_pixel_value}")
    for chunk in iterate_chunks(output_array.shape, max_chunk_size=256 * 1024**2):
        output_array[chunk][output_footprint[chunk] == 0] = blank_pixel_value

    return output_array, output_footprint
